/*
 *  A FLAC decoder plugin for the Beep Media Player
 *  Copyright (C) 2005 Ralf Ertzinger
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#ifndef FLAC_H
#define FLAC_H

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <FLAC/all.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <bmp/plugin.h>
#include <bmp/vfs.h>
#include <bmp/util.h>
#include <bmp/titlestring.h>
#include "debug.h"
#include "version.h"
#include "seekable_stream_callbacks.h"

#define OUTPUT_BLOCK_SIZE (8192u)
#define MAX_SUPPORTED_CHANNELS (2u)
#define BUFFER_SIZE (FLAC__MAX_BLOCK_SIZE * MAX_SUPPORTED_CHANNELS)

struct stream_info {
    guint bits_per_sample;
    guint samplerate;
    guint channels;
    gulong samples;
};

struct frame_info {
    guint bits_per_sample;
    guint samplerate;
    guint channels;
};

struct stream_comment {
    gchar* artist;
    gchar* album;
    gchar* title;
    gchar* tracknumber;
    gchar* genre;
};

typedef struct callback_info {
    gint16* output_buffer;
    gint16* write_pointer;
    guint buffer_free;
    guint buffer_used;
    VFSFile* input_stream;
    struct stream_info stream;
    struct stream_comment comment;
    gboolean metadata_changed;
    struct frame_info frame;
    glong read_max;
} callback_info;


void flac_init(void);
void flac_aboutbox(void);
gint flac_is_our_file(gchar *filename);
void flac_play_file(gchar *filename);
void flac_stop(void);
void flac_seek(gint time);
void flac_pause(gshort p);
gint flac_get_time(void);
void flac_get_song_info(gchar *filename, gchar **title, gint *length);

callback_info* init_callback_info(void);
void reset_info(callback_info* info);
gchar* SeekableStreamDecoderState(FLAC__SeekableStreamDecoderState state);
gchar* StreamDecoderState(FLAC__StreamDecoderState state);
gboolean read_metadata(gchar* filename, FLAC__SeekableStreamDecoder* decoder, callback_info* info);
gchar* get_title(const gchar* filename, callback_info* info);
void add_comment(callback_info* info, gchar* key, gchar* value);

#endif
