/*
 *  A FLAC decoder plugin for the Beep Media Player
 *  Copyright (C) 2005 Ralf Ertzinger
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "flac.h"

/* === */

callback_info* init_callback_info(void) {

    callback_info* info;

    _ENTER;

    if (NULL == (info = malloc(sizeof(callback_info)))) {
        _ERROR("Could not allocate memory for callback structure!");
        _LEAVE NULL;
    }

    if (NULL == (info->output_buffer = malloc(BUFFER_SIZE * sizeof(int16_t)))) {
        _ERROR("Could not allocate memory for output buffer!");
        _LEAVE NULL;
    }

    /*
     * We need to set this manually to NULL because
     * reset_info will try to close the file pointed to
     * by input_stream if it is non-NULL.
     * Same for info->comment.x
     */
    info->input_stream = NULL;
    info->comment.artist = NULL;
    info->comment.album = NULL;
    info->comment.title = NULL;
    info->comment.tracknumber = NULL;
    info->comment.genre = NULL;
    reset_info(info);

    _DEBUG("Playback buffer allocated for %d elements, %d bytes each", BUFFER_SIZE, sizeof(int16_t));

    _LEAVE info;
}

/* --- */

void reset_info(callback_info* info) {

    _ENTER;

    _DEBUG("Using callback_info at %p", info);

    if (NULL != info->input_stream) {
        vfs_fclose(info->input_stream);
        info->input_stream = NULL;
    }

    // memset(info->output_buffer, 0, BUFFER_SIZE * sizeof(int16_t));
    info->stream.samplerate = 0;
    info->stream.bits_per_sample = 0;
    info->stream.channels = 0;
    info->stream.samples = 0;
    info->buffer_free = BUFFER_SIZE;
    info->buffer_used = 0;
    info->write_pointer = info->output_buffer;
    info->read_max = -1;

    /*
     * Clear the stream and frame information
     */
    info->stream.bits_per_sample = 0;
    info->stream.samplerate = 0;
    info->stream.channels = 0;
    info->stream.samples = 0;

    if (NULL != info->comment.artist) {
        free(info->comment.artist);
        info->comment.artist = NULL;
    }

    if (NULL != info->comment.album) {
        free(info->comment.album);
        info->comment.album = NULL;
    }

    if (NULL != info->comment.title) {
        free(info->comment.title);
        info->comment.title = NULL;
    }

    if (NULL != info->comment.tracknumber) {
        free(info->comment.tracknumber);
        info->comment.tracknumber = NULL;
    }

    if (NULL != info->comment.genre) {
        free(info->comment.genre);
        info->comment.genre = NULL;
    }

    info->frame.bits_per_sample = 0;
    info->frame.samplerate = 0;
    info->frame.channels = 0;

    info->metadata_changed = FALSE;

    _LEAVE;
}

/* --- */

gchar* SeekableStreamDecoderState(FLAC__SeekableStreamDecoderState state) {

    _ENTER;

    switch(state) {

        case FLAC__SEEKABLE_STREAM_DECODER_OK:
            _LEAVE "FLAC__SEEKABLE_STREAM_DECODER_OK";
            break;

        case FLAC__SEEKABLE_STREAM_DECODER_SEEKING:
            _LEAVE "FLAC__SEEKABLE_STREAM_DECODER_SEEKING";
            break;

        case FLAC__SEEKABLE_STREAM_DECODER_END_OF_STREAM:
            _LEAVE "FLAC__SEEKABLE_STREAM_DECODER_END_OF_STREAM";
            break;

        case FLAC__SEEKABLE_STREAM_DECODER_MEMORY_ALLOCATION_ERROR:
            _LEAVE "FLAC__SEEKABLE_STREAM_DECODER_MEMORY_ALLOCATION_ERROR";
            break;

        case FLAC__SEEKABLE_STREAM_DECODER_STREAM_DECODER_ERROR:
            _LEAVE "FLAC__SEEKABLE_STREAM_DECODER_STREAM_DECODER_ERROR";
            break;

        case FLAC__SEEKABLE_STREAM_DECODER_READ_ERROR:
            _LEAVE "FLAC__SEEKABLE_STREAM_DECODER_READ_ERROR";
            break;

        case FLAC__SEEKABLE_STREAM_DECODER_SEEK_ERROR:
            _LEAVE "FLAC__SEEKABLE_STREAM_DECODER_SEEK_ERROR";
            break;

        case FLAC__SEEKABLE_STREAM_DECODER_ALREADY_INITIALIZED:
            _LEAVE "FLAC__SEEKABLE_STREAM_DECODER_ALREADY_INITIALIZED";
            break;

        case FLAC__SEEKABLE_STREAM_DECODER_INVALID_CALLBACK:
            _LEAVE "FLAC__SEEKABLE_STREAM_DECODER_INVALID_CALLBACK";
            break;

        case FLAC__SEEKABLE_STREAM_DECODER_UNINITIALIZED:
            _LEAVE "FLAC__SEEKABLE_STREAM_DECODER_UNINITIALIZED";
            break;

    }

    _LEAVE "";
}

/* --- */

gchar* StreamDecoderState(FLAC__StreamDecoderState state) {

    _ENTER;

    switch(state) {

        case FLAC__STREAM_DECODER_SEARCH_FOR_METADATA:
            _LEAVE "FLAC__STREAM_DECODER_SEARCH_FOR_METADATA";
            break;

        case FLAC__STREAM_DECODER_READ_METADATA:
            _LEAVE "FLAC__STREAM_DECODER_READ_METADATA";
            break;

        case FLAC__STREAM_DECODER_SEARCH_FOR_FRAME_SYNC:
            _LEAVE "FLAC__STREAM_DECODER_SEARCH_FOR_FRAME_SYNC";
            break;

        case FLAC__STREAM_DECODER_READ_FRAME:
            _LEAVE "FLAC__STREAM_DECODER_READ_FRAME";
            break;

        case FLAC__STREAM_DECODER_END_OF_STREAM:
            _LEAVE "FLAC__STREAM_DECODER_END_OF_STREAM";
            break;

        case FLAC__STREAM_DECODER_ABORTED:
            _LEAVE "FLAC__STREAM_DECODER_ABORTED";
            break;

        case FLAC__STREAM_DECODER_UNPARSEABLE_STREAM:
            _LEAVE "FLAC__STREAM_DECODER_UNPARSEABLE_STREAM";
            break;

        case FLAC__STREAM_DECODER_MEMORY_ALLOCATION_ERROR:
            _LEAVE "FLAC__STREAM_DECODER_MEMORY_ALLOCATION_ERROR";
            break;

        case FLAC__STREAM_DECODER_ALREADY_INITIALIZED:
            _LEAVE "FLAC__STREAM_DECODER_ALREADY_INITIALIZED";
            break;

        case FLAC__STREAM_DECODER_INVALID_CALLBACK:
            _LEAVE "FLAC__STREAM_DECODER_INVALID_CALLBACK";
            break;

        case FLAC__STREAM_DECODER_UNINITIALIZED:
            _LEAVE "FLAC__STREAM_DECODER_UNINITIALIZED";
            break;

    }

    _LEAVE "";
}

/* --- */

gboolean read_metadata(gchar* filename, FLAC__SeekableStreamDecoder* decoder, callback_info* info) {

    FLAC__SeekableStreamDecoderState ret;

    _ENTER;

    _DEBUG("Using callback_info at %p", info);

    _DEBUG("Opening file %s", filename);

    /*
     * Reset the decoder
     */
    if (false == FLAC__seekable_stream_decoder_reset(decoder)) {
        _ERROR("Could not reset the decoder!");
        _LEAVE FALSE;
    }

    reset_info(info);

    /*
     * Just scan the first 8k for the start of metadata
     */

    info->read_max = 8192;

    /*
     * Open the file
     */
    if (NULL == (info->input_stream = vfs_fopen(filename, "rb"))) {
        _ERROR("Could not open file for reading!");
        _LEAVE FALSE;
    }

    /*
     * Try to decode the metadata
     */
    if (false == FLAC__seekable_stream_decoder_process_until_end_of_metadata(decoder)) {
        ret = FLAC__seekable_stream_decoder_get_state(decoder);
        _DEBUG("Could not read the metadata: %s(%d)!", SeekableStreamDecoderState(ret), ret);
        reset_info(info);
        _LEAVE FALSE;
    }

    /*
     * Seems to be a FLAC stream. Allow further reading
     */

    info->read_max = -1;

    _LEAVE TRUE;
}

/* --- */

static const gchar* get_extension(const gchar* filename) {

    _ENTER;

    const gchar *ext = strrchr(filename, '.');
    _LEAVE ext ? ext + 1 : NULL;
}

/* --- */

gchar* get_title(const gchar* filename, callback_info* info) {

    TitleInput *input;
    gchar *title;

    _ENTER;

    _DEBUG("Using callback_info at %p", info);

    input = bmp_title_input_new();

    input->file_name = g_path_get_basename(filename);
    input->file_ext = get_extension(filename);
    input->file_path = g_path_get_dirname(filename);

    input->performer = g_strdup(info->comment.artist);
    input->track_name = g_strdup(info->comment.title);
    input->album_name = g_strdup(info->comment.album);

    if (!(title = xmms_get_titlestring(xmms_get_gentitle_format(), input)))
        title = g_strdup(input->file_name);

    bmp_title_input_free(input);

    _DEBUG("Title created: <%s>", title);

    _LEAVE title;
}

/* --- */

void add_comment(callback_info* info, gchar* key, gchar* value) {

    /*
     * We are currently interested in comments of type
     * - ARTIST
     * - ALBUM
     * - TITLE
     * - TRACKNUMBER
     *
     * We have to create copies of the value string.
     */

    gchar** destination = NULL;

     _ENTER;

     _DEBUG("Using callback_info at %p", info);

    if (0 == strcasecmp(key, "ARTIST")) {
        _DEBUG("Found key ARTIST");
        destination = &(info->comment.artist);
    }

    if (0 == strcasecmp(key, "ALBUM")) {
        _DEBUG("Found key ALBUM");
        destination = &(info->comment.album);
    }

    if (0 == strcasecmp(key, "TITLE")) {
        _DEBUG("Found key TITLE");
        destination = &(info->comment.title);
    }

    if (0 == strcasecmp(key, "TRACKNUMBER")) {
        _DEBUG("Found key TRACKNUMBER");
        destination = &(info->comment.tracknumber);
    }

    if (NULL != destination) {
        if (NULL != *destination) {
            g_free(*destination);
        }

        if (NULL == (*destination = g_strdup(value))) {
            _ERROR("Could not allocate memory for comment!");
            _LEAVE;
        }
    }

    _LEAVE;
}
